#include "viewer.h"
#include <GL/glu.h>
#include "mesh.h"
#include "gcoder.h"
#include <QMouseEvent>
#include <QWheelEvent>
#include "utils.h"
#include <QGLFormat>
#include <iostream>
#include "printsettings.h"
#include <utility>

Viewer *Viewer::pInstance = NULL;
Viewer *Viewer::instance()
{
    if (!pInstance)
    {
        QGLFormat fmt = QGLFormat::defaultFormat();
        fmt.setSampleBuffers(false);
        fmt.setDirectRendering(true);
        fmt.setDoubleBuffer(true);
        fmt.setSamples(4);
        pInstance = new Viewer(fmt);
    }
    return pInstance;
}

Viewer::Viewer(const QGLFormat &fmt) :
    QGLWidget(fmt)
{
    mesh = NULL;
    lower = Vec3(0,0,0);
    upper = Vec3(180,180,140);
}

Viewer::~Viewer()
{
    if (mesh)
        delete mesh;
}

void Viewer::setMesh(Mesh *mesh)
{
    if (this->mesh == mesh)
        return;
    if (this->mesh)
        delete this->mesh;
    this->mesh = mesh;

    bbox = mesh->getBoundingBox();
	mesh->transform(Mat44::translation(-bbox.first));
	bbox = mesh->getBoundingBox();
	mesh->backup();

	PrintSettings::instance()->resetTransform();

    makeCurrent();
    initializeGL();
    doneCurrent();

    update();
}

void Viewer::setBuildRange(const Vec3 &lower, const Vec3 &upper)
{
    this->lower = lower;
    this->upper = upper;

    makeCurrent();
    initializeGL();
    doneCurrent();

    update();
}

void Viewer::initializeGL()
{
    glClearColor(0,0,0,0);
    glDisable(GL_TEXTURE_2D);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30.0, double(width()) / qMax(1.0, double(height())), 1.0, 1.0e4);
    glMatrixMode(GL_MODELVIEW);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    Vec3 center = 0.5 * (lower + upper);
    Vec3 eye;
    if (mesh)
    {
		center = 0.5 * (bbox.first + bbox.second);
        eye = center + 2.0 * Vec3(0,-(bbox.second.y() - bbox.first.y()),bbox.second.z() - bbox.first.z());
    }
    else
        eye = center + 2.0 * Vec3(0, -(upper.y() - lower.y()), upper.z() - lower.z());
    const Vec3 up(0,0,1);
    glLoadIdentity();
    gluLookAt(eye.x(), eye.y(), eye.z(), center.x(), center.y(), center.z(), up.x(), up.y(), up.z());
    model_matrix = Mat44::identity();
}

void Viewer::resizeGL(int w, int h)
{
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30.0, double(w) / qMax(1.0, double(h)), 1.0, 1.0e4);
    glMatrixMode(GL_MODELVIEW);
}

void Viewer::paintGL()
{
    glPushMatrix();
    glMultMatrixd((const GLdouble*)model_matrix.data());
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glDisable(GL_BLEND);
    glEnable(GL_COLOR_MATERIAL);
    glColor3ub(0xFF,0xFF,0xFF);
    drawBox(lower, upper);

    glColor3ub(0xFF,0xFF,0);
    glBegin(GL_LINES);
    for(int x = std::ceil(lower.x()) ; x <= upper.x() ; x += 10)
    {
        glVertex3f(x, lower.y(), lower.z());
        glVertex3f(x, upper.y(), lower.z());
    }
    for(int y = std::ceil(lower.y()) ; y <= upper.y() ; y += 10)
    {
        glVertex3f(lower.x(), y, lower.z());
        glVertex3f(upper.x(), y, lower.z());
    }
    glEnd();

    if (mesh)
    {
        glPushMatrix();

        glColor3ub(0xFF,0,0);
        drawBox(bbox.first, bbox.second);

        glColor3ub(0,0,0xFF);
        GCoder::instance()->draw();

        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
        glColor4ub(0xFF,0xFF,0xFF,0x7F);
        glEnable(GL_LIGHTING);
        const GLfloat light_pos[4] = { 1.0f, 1.0f, 1.0f, 0.0f };
        const GLfloat light_diffuse[4] = { 1.0f, 1.0f, 1.0f, 0.0f };
        const GLfloat light_specular[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
        const GLfloat light_ambient[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
        glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
        glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
        glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
        glEnable(GL_LIGHT0);
        glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
		mesh->draw();
        glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
        mesh->draw();
        glDisable(GL_BLEND);
        glDisable(GL_LIGHTING);
        glPopMatrix();
    }
    glPopMatrix();
}

void Viewer::drawBox(const Vec3 &lower, const Vec3 &upper) const
{
    glBegin(GL_LINES);
    glVertex3f(lower.x(), lower.y(), lower.z());    glVertex3f(upper.x(), lower.y(), lower.z());
    glVertex3f(lower.x(), upper.y(), lower.z());    glVertex3f(upper.x(), upper.y(), lower.z());
    glVertex3f(lower.x(), lower.y(), upper.z());    glVertex3f(upper.x(), lower.y(), upper.z());
    glVertex3f(lower.x(), upper.y(), upper.z());    glVertex3f(upper.x(), upper.y(), upper.z());

    glVertex3f(lower.x(), lower.y(), lower.z());    glVertex3f(lower.x(), upper.y(), lower.z());
    glVertex3f(upper.x(), lower.y(), lower.z());    glVertex3f(upper.x(), upper.y(), lower.z());
    glVertex3f(lower.x(), lower.y(), upper.z());    glVertex3f(lower.x(), upper.y(), upper.z());
    glVertex3f(upper.x(), lower.y(), upper.z());    glVertex3f(upper.x(), upper.y(), upper.z());

    glVertex3f(lower.x(), lower.y(), lower.z());    glVertex3f(lower.x(), lower.y(), upper.z());
    glVertex3f(upper.x(), lower.y(), lower.z());    glVertex3f(upper.x(), lower.y(), upper.z());
    glVertex3f(lower.x(), upper.y(), lower.z());    glVertex3f(lower.x(), upper.y(), upper.z());
    glVertex3f(upper.x(), upper.y(), lower.z());    glVertex3f(upper.x(), upper.y(), upper.z());
    glEnd();
}

void Viewer::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton)
        last_cursor_pos = e->pos();
}

void Viewer::mouseMoveEvent(QMouseEvent *e)
{
    if (e->buttons() == Qt::LeftButton)
    {
        Vec3 center = 0.5 * (lower + upper);
        double radius = (upper - lower).length() * 0.5;
        if (mesh)
        {
			center = 0.5 * (bbox.first + bbox.second);
            radius = (bbox.second - bbox.first).length() * 0.5;
        }

        makeCurrent();
        Mat44 K;
        Mat44 viewmatrix;
        glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrix.data());
        glGetDoublev(GL_PROJECTION_MATRIX, K.data());
        const Mat33 IK = K.sub(3,3).transposed().inverse();

        const Mat44 IM = viewmatrix.transposed().inverse();
        const Vec3 eye = IM * Vec3(0,0,0);
        const Mat33 IM3 = IM.sub(3,3);
        Vec3 dir0 = IM3 * (IK * Vec3(double(last_cursor_pos.x()) / width() * 2.0 - 1.0,
                                     -(double(last_cursor_pos.y()) / height() * 2.0 - 1.0),
                                     -1.0));
        Vec3 dir1 = IM3 * (IK * Vec3(double(e->x()) / width() * 2.0 - 1.0,
                                     -(double(e->y()) / height() * 2.0 - 1.0),
                                     -1.0));
        dir0.normalize();
        dir1.normalize();

        Vec3 hit_point0;
        if (!hit_sphere(eye, dir0, center, radius, hit_point0))
        {
            doneCurrent();
            last_cursor_pos = e->pos();
            return;
        }
        Vec3 hit_point1;
        if (!hit_sphere(eye, dir1, center, radius, hit_point1))
        {
            doneCurrent();
            last_cursor_pos = e->pos();
            return;
        }

        dir0 = hit_point0 - center;
        dir1 = hit_point1 - center;
        dir0.normalize();
        dir1.normalize();
        Vec3 u = dir0 ^ dir1;
        u.normalize();

        glLoadIdentity();
        glTranslated(center.x(), center.y(), center.z());
        glRotated(std::acos(dir0 | dir1) * 180.0 / M_PI, u.x(), u.y(), u.z());
        glTranslated(-center.x(),-center.y(),-center.z());
        glMultMatrixd((const GLdouble*)model_matrix.data());
        glGetDoublev(GL_MODELVIEW_MATRIX, (GLdouble*)model_matrix.data());

        Vec3 i(model_matrix(0,0), model_matrix(1,0), model_matrix(2,0));
        Vec3 j(model_matrix(0,1), model_matrix(1,1), model_matrix(2,1));
        Vec3 k(model_matrix(0,2), model_matrix(1,2), model_matrix(2,2));
        const double s = (i.length() + j.length() + k.length()) / 3.0;
        k = i ^ j;
        i.normalize();
        i = i * s;
        j.normalize();
        j = j * s;
        k.normalize();
        k = k * s;
        for(int l = 0 ; l < 3 ; ++l)
        {
            model_matrix(l,0) = i[l];
            model_matrix(l,1) = j[l];
            model_matrix(l,2) = k[l];
        }

        glLoadMatrixd((const GLdouble*)viewmatrix.data());

        doneCurrent();
        last_cursor_pos = e->pos();

        updateGL();
    }
}

void Viewer::wheelEvent(QWheelEvent *e)
{
    if (e->delta() != 0)
    {
        Vec3 center = 0.5 * (lower + upper);
        if (mesh)
			center = 0.5 * (bbox.first + bbox.second);

        makeCurrent();
        GLdouble tmp[16];
        glGetDoublev(GL_MODELVIEW_MATRIX, tmp);
        glLoadIdentity();

        glTranslated(center.x(), center.y(), center.z());
		const double s = std::exp(3.0e-4 * e->delta());
        glScaled(s, s, s);
        glTranslated(-center.x(),-center.y(),-center.z());
        glMultMatrixd((const GLdouble*)model_matrix.data());
        glGetDoublev(GL_MODELVIEW_MATRIX, (GLdouble*)model_matrix.data());

        glLoadMatrixd((const GLdouble*)tmp);

        doneCurrent();

        updateGL();
    }
}

void Viewer::transformMesh(const Mat44 &m)
{
	if (mesh)
	{
		mesh->transform(m);
		bbox = mesh->getBoundingBox();
		makeCurrent();
		initializeGL();
		doneCurrent();
		updateGL();
	}
}

std::pair<Vec3, Vec3> Viewer::getBuildRange() const
{
	return std::make_pair(lower, upper);
}

void Viewer::invertNormals()
{
	if (mesh)
	{
		mesh->invertNormals();
		updateGL();
	}
}
